/* drvHMS.cpp

    Author: Mark Rivers

    This is the driver for the Acromag HMSA digital input IP module
*/

#include <string.h>

/* EPICS includes */
#include <errlog.h>
#include <epicsTypes.h>
#include <epicsExport.h>
#include <epicsThread.h>
#include <epicsString.h>
#include <iocsh.h>

#include <asynPortDriver.h>
#include <asynOctetSyncIO.h>

#define MAX_SEND_MESSAGE    8
#define MAX_RECEIVE_MESSAGE 1034
#define HEADER_LEN          4
#define MESSAGE_VERSION     1
#define NUM_ADCS            12
#define NUM_WAVEFORM_ADCS   8
#define NUM_TRIGGERS        17
#define NUM_LONG_REGISTERS  4
#define NUM_SHORT_REGISTERS 4
#define MAX_ADDRESS         17
#define WAVEFORM_LEN        512

#define ADC_RESPONSE_LEN      28
#define REGISTER_RESPONSE_LEN 28
#define TRIGGER_RESPONSE_LEN  38
#define WAVEFORM_RESPONSE_LEN 1032

/* Message codes */
#define TRIGGER_ENABLE_CODE     0x07
#define TRIGGER_DISABLE_CODE    0x08
#define READ_REGISTERS_CODE     0x41
#define WRITE_REGISTER_CODE     0x11
#define READ_ADCS_CODE          0x48
#define READ_TRIGGERS_CODE      0x4b
#define WRITE_TRIGGER_CODE      0x13
#define READ_WAVEFORMS_CODE     0x51

#define TRIGGER_ENABLE_STRING           "TRIGGER_ENABLE"
#define UNSOLICITED_MESSAGES_STRING     "UNSOLICITED_MESSAGES"
#define READ_REGISTERS_STRING           "READ_REGISTERS"
#define READ_ADCS_STRING                "READ_ADCS"
#define READ_TRIGGERS_STRING            "READ_TRIGGERS"
#define READ_WAVEFORMS_STRING           "READ_WAVEFORMS"
#define REGISTER_VALUE_STRING           "REGISTER_VALUE"
#define ADC_VALUE_STRING                "ADC_VALUE"
#define TRIGGER_VALUE_STRING            "TRIGGER_VALUE"
#define WAVEFORM_VALUE_STRING           "WAVEFORM_VALUE"

#define TRIGGER_TIMEOUT .01
#define WRITE_READ_TIMEOUT 1.0

static const char *driverName = "HMS";

/** This is the class definition for the HMS class*/
class HMS : public asynPortDriver
{
public:
    HMS(const char *portName, const char *HMSPortName, int HMId);
    virtual asynStatus writeInt32(asynUser *pasynUser, epicsInt32 value);
    virtual asynStatus readInt16Array(asynUser *pasynUser, epicsInt16 *value, size_t nElements, size_t *nIn);
    virtual void report(FILE *fp, int details);
    // These should be private, but are called from C, so must be public
    void triggerThread();  

private:
    asynStatus writeReadHMS(int messageCode, epicsInt16 *data, size_t dataLen, size_t *responseLen);
    asynStatus readRegisters();
    asynStatus readADCs();
    asynStatus readTriggers();
    asynStatus readWaveforms();

    // These contain the parameter indices for each of the driver parameters
    int triggerEnableParam;
    #define FIRST_HMS_PARAM triggerEnableParam
    int unsolicitedMessagesParam;
    int readRegistersParam;
    int readADCsParam;
    int readTriggersParam;
    int readWaveformsParam;
    int registerValueParam;
    int ADCValueParam;
    int triggerValueParam;
    int waveformValueParam;

    #define LAST_HMS_PARAM waveformValueParam
    char *HMSServerName;
    int managerID;
    asynUser *pasynUserHMS;
    char sendMessage[MAX_SEND_MESSAGE];
    char receiveMessage[MAX_RECEIVE_MESSAGE];
    epicsInt16 waveforms[WAVEFORM_LEN][NUM_WAVEFORM_ADCS];
};

#define NUM_HMS_PARAMS (&LAST_HMS_PARAM - &FIRST_HMS_PARAM + 1)

// These functions must have C linkage because they are called from other EPICS components
extern "C" {
static void triggerThreadC(void * pPvt)
{
    HMS *pHMS = (HMS *)pPvt;
    pHMS->triggerThread();
}}

HMS::HMS(const char *portName, const char *HMSPortName, int HMId)
    :asynPortDriver(portName, MAX_ADDRESS, NUM_HMS_PARAMS,                // portName, maxAddr, paramTableSize
                    asynInt32Mask | asynInt16ArrayMask | asynDrvUserMask, // interfaceMask
                    asynInt32Mask | asynInt16ArrayMask,                   // interruptMask
                    ASYN_CANBLOCK | ASYN_MULTIDEVICE, 1, 0, 0)            // asynFlags, autoConnect, priority, stackSize
{
    static const char *functionName = "HMS";
    asynStatus status;
    
    managerID = HMId;
    HMSServerName = epicsStrDup(HMSPortName);

    /* Connect to HMS server */
    status = pasynOctetSyncIO->connect(HMSPortName, 0, &this->pasynUserHMS, NULL);
    if (status != asynSuccess) {
        asynPrint(pasynUserSelf, ASYN_TRACE_ERROR, 
            "%s:%s: cannot connect to HMS server\n",
            driverName, functionName);
    }

    /* Create the asynPortDriver parameters */
    createParam(TRIGGER_ENABLE_STRING,          asynParamInt32, &triggerEnableParam); 
    createParam(UNSOLICITED_MESSAGES_STRING,    asynParamInt32, &unsolicitedMessagesParam); 
    createParam(READ_REGISTERS_STRING,          asynParamInt32, &readRegistersParam); 
    createParam(READ_ADCS_STRING,               asynParamInt32, &readADCsParam); 
    createParam(READ_TRIGGERS_STRING,           asynParamInt32, &readTriggersParam); 
    createParam(READ_WAVEFORMS_STRING,          asynParamInt32, &readWaveformsParam); 
    createParam(REGISTER_VALUE_STRING,          asynParamInt32, &registerValueParam); 
    createParam(ADC_VALUE_STRING,               asynParamInt32, &ADCValueParam); 
    createParam(TRIGGER_VALUE_STRING,           asynParamInt32, &triggerValueParam); 
    createParam(WAVEFORM_VALUE_STRING,          asynParamInt16Array, &waveformValueParam); 

    /* Create the send message header */
    sendMessage[0] = MESSAGE_VERSION;
    sendMessage[1] = managerID;
    sendMessage[3] = 0;
    
    /* Read the current values of the hardware registers so output records will be initialized correctly */
    readRegisters();
    readADCs();
    readTriggers();
    readWaveforms();
    
    /* Start the thread to poll and handle interrupt callbacks to 
     * device support */
    epicsThreadCreate("HMS",
                      epicsThreadPriorityMedium,
                      epicsThreadGetStackSize(epicsThreadStackMedium),
                      (EPICSTHREADFUNC)triggerThreadC,
                      this);
}

asynStatus HMS::writeReadHMS(int messageCode, epicsInt16 *data, size_t dataLen, size_t *responseLen)
{
    size_t numSend, numSent;
    int eomReason;
    asynStatus status;
    
    numSend = HEADER_LEN + dataLen;
    if (dataLen) memcpy(&sendMessage[4], data, dataLen);
    sendMessage[2] = messageCode;
    status = pasynOctetSyncIO->writeRead(pasynUserHMS, sendMessage, numSend, 
                                         receiveMessage, sizeof(receiveMessage), 
                                         WRITE_READ_TIMEOUT, &numSent, responseLen, &eomReason);
    return(status);
}    

asynStatus HMS::readRegisters()
{
    static const char *functionName = "readRegisters";
    size_t responseLen;
    asynStatus status;
    int i;
    epicsInt32 *longRegVals  = (epicsInt32 *)&receiveMessage[4];
    epicsInt16 *shortRegVals = (epicsInt16 *)&receiveMessage[20];

    status = writeReadHMS(READ_REGISTERS_CODE, NULL, 0, &responseLen);
    if ((status != asynSuccess) || (responseLen != REGISTER_RESPONSE_LEN)) {
        asynPrint(pasynUserSelf, ASYN_TRACE_ERROR, 
            "%s:%s: error reading HMS registers status=%d response length=%d, expected %d\n",
            driverName, functionName, status, responseLen, REGISTER_RESPONSE_LEN);
        return(asynError);
    }
    // Copy the long register values
    for (i=0; i<NUM_LONG_REGISTERS; i++) setIntegerParam(i, registerValueParam, longRegVals[i]);
    // Copy the short register values
    for (i=0; i<NUM_SHORT_REGISTERS; i++) setIntegerParam(i+NUM_LONG_REGISTERS, registerValueParam, shortRegVals[i]);
    return(asynSuccess);
}

asynStatus HMS::readADCs()
{
    static const char *functionName = "readADCs";
    size_t responseLen;
    asynStatus status;
    int i;
    epicsInt16 *ADCVals = (epicsInt16 *)&receiveMessage[4];

    status = writeReadHMS(READ_ADCS_CODE, NULL, 0, &responseLen);
    if ((status != asynSuccess) || (responseLen != ADC_RESPONSE_LEN)) {
        asynPrint(pasynUserSelf, ASYN_TRACE_ERROR, 
            "%s:%s: error reading HMS ADCs status=%d response length=%d, expected %d\n",
            driverName, functionName, status, responseLen, ADC_RESPONSE_LEN);
        return(asynError);
    }
    // Copy the ADC values
    for (i=0; i<NUM_ADCS; i++) setIntegerParam(i, ADCValueParam, ADCVals[i]);
    return(asynSuccess);
}

asynStatus HMS::readTriggers()
{
    static const char *functionName = "readTriggers";
    size_t responseLen;
    asynStatus status;
    int i;
    epicsInt16 *triggerVals = (epicsInt16 *)&receiveMessage[4];

    status = writeReadHMS(READ_TRIGGERS_CODE, NULL, 0, &responseLen);
    if ((status != asynSuccess) || (responseLen != TRIGGER_RESPONSE_LEN)) {
        asynPrint(pasynUserSelf, ASYN_TRACE_ERROR, 
            "%s:%s: error reading HMS triggers status=%d response length=%d, expected %d\n",
            driverName, functionName, status, responseLen, TRIGGER_RESPONSE_LEN);
        return(asynError);
    }
    // Copy the trigger values
    for (i=0; i<NUM_TRIGGERS; i++) setIntegerParam(i, triggerValueParam, triggerVals[i]);
    return(asynSuccess);
}

asynStatus HMS::readWaveforms()
{
    static const char *functionName = "readWaveforms";
    size_t responseLen;
    asynStatus status;
    int i;
    epicsInt16 sendData[2];

    for (i=0; i<NUM_WAVEFORM_ADCS; i++) {
        sendData[0] = i;  // ADC number
        sendData[1] = 0;  // Block number
        status = writeReadHMS(READ_WAVEFORMS_CODE, sendData, sizeof(sendData), &responseLen);
        if ((status != asynSuccess) || (responseLen != WAVEFORM_RESPONSE_LEN)) {
            asynPrint(pasynUserSelf, ASYN_TRACE_ERROR, 
                "%s:%s: error reading HMS waveforms status=%d response length=%d, expected %d\n",
                driverName, functionName, status, responseLen, WAVEFORM_RESPONSE_LEN);
            return(asynError);
        }
        // Copy the waveform to object
        memcpy(&waveforms[0][i], &receiveMessage[8], WAVEFORM_LEN*sizeof(epicsInt16));
        doCallbacksInt16Array(&waveforms[0][i], WAVEFORM_LEN, waveformValueParam, i);
    }
    return(asynSuccess);
}

asynStatus HMS::writeInt32(asynUser *pasynUser, epicsInt32 value)
{
    static const char *functionName = "writeInt32";
    int function = pasynUser->reason;
    int code;
    int address;
    int i;
    size_t responseLen;
    asynStatus status = asynSuccess;
    epicsInt16 data[2];
    
    getAddress(pasynUser, &address);
    setIntegerParam(address, function, value);
    
    if (function == triggerEnableParam) {
        code = (value) ? TRIGGER_ENABLE_CODE : TRIGGER_DISABLE_CODE;
        status = writeReadHMS(code, NULL, 0, &responseLen);
    } 
    else if (function == readRegistersParam) {
        status = readRegisters();
    }
    else if (function == readADCsParam) {
        status = readADCs();
    }
    else if (function == readTriggersParam) {
        status = readTriggers();
    }
    else if (function == readWaveformsParam) {
        status = readWaveforms();
    }
    else if (function == registerValueParam) {
        data[0] = address-NUM_LONG_REGISTERS;
        data[1] = value;
        status = writeReadHMS(WRITE_REGISTER_CODE, data, 4, &responseLen);
    }
    else if (function == triggerValueParam) {
        data[0] = address;
        data[1] = value;
        status = writeReadHMS(WRITE_TRIGGER_CODE, data, 4, &responseLen);
    }
    asynPrint(pasynUser, ASYN_TRACEIO_DRIVER,
              "%s:%s:, function=%d, address=%d, value=%d, status=%d\n", 
              driverName, functionName, function, address, value, status);
    for (i=0; i<MAX_ADDRESS; i++) callParamCallbacks(i);
    return(status);
}

asynStatus HMS::readInt16Array(asynUser *pasynUser, epicsInt16 *value, size_t nElements, size_t *nIn)
{
    int function = pasynUser->reason;
    int address;
    asynStatus status = asynError;

    getAddress(pasynUser, &address);
    if (function == waveformValueParam) {
        *nIn = WAVEFORM_LEN;
        if (*nIn > nElements) {
            *nIn = nElements;
        }
        /* Copy the data */
        memcpy(value, &waveforms[0][address], *nIn * sizeof(epicsInt16));
        status = asynSuccess;
    }
    return(status);
}

void HMS::triggerThread()
{
    /* This function runs in a separate thread. */
    static const char *functionName = "triggerThread";
    char message[4];
    asynStatus status;
    int count;
    size_t len;
    int eomReason;

    setIntegerParam(unsolicitedMessagesParam, 0);
    
    while(1) {      
        status = pasynOctetSyncIO->read(pasynUserHMS, message, sizeof(message), TRIGGER_TIMEOUT, &len, &eomReason);
        if (status == asynSuccess) {
            getIntegerParam(unsolicitedMessagesParam, &count);
            count++;
            setIntegerParam(unsolicitedMessagesParam, count);
            callParamCallbacks();
            asynPrint(this->pasynUserSelf, ASYN_TRACEIO_DRIVER,
                        "%s:%s:, got unsolicited message\n", 
                        driverName, functionName);
        }
    }
}

void HMS::report(FILE *fp, int details)
{
    fprintf(fp, "%s: Connected to HMS server %s ID=%d\n",
        portName, HMSServerName, managerID);
    asynPortDriver::report(fp, details);
}

extern "C" int initHMS(const char *portName, const char *HMSPortName, int HMId)
{
    HMS *pHMS = new HMS(portName, HMSPortName, HMId);
    pHMS = NULL;
    return(asynSuccess);
}

/* iocsh functions */
static const iocshArg initArg0 = { "Port name",iocshArgString};
static const iocshArg initArg1 = { "HMS port name",iocshArgString};
static const iocshArg initArg2 = { "HM ID",iocshArgInt};
static const iocshArg * const initArgs[3] = {&initArg0,
                                             &initArg1,
                                             &initArg2};

static const iocshFuncDef initFuncDef = {"initHMS",3,initArgs};

static void initCallFunc(const iocshArgBuf *args)
{
    initHMS(args[0].sval, args[1].sval, args[2].ival);
}

void HMSRegister(void)
{
    iocshRegister(&initFuncDef,initCallFunc);
}

epicsExportRegistrar(HMSRegister);






